(* Copyright (c) 2021 CIIRC (Czech Institute of Informatics, Robotics and Cybernetics) / CTU (Czech Technical University) *)

open Syntax
open Parser
open Sha256
open Hash

(*** The hash ids for the following will be added as we bootstrap, allowing us to make sense of more term syntax as we go. ***)
let debug = false;;
let fal : string ref = ref "";;
let fale : string ref = ref "";;
let eqPoly : string ref = ref "";;
let eqsymPoly : string ref = ref "";;
let eqsymPolyknown : bool ref = ref false;;
let eqSet : string ref = ref "";;
let conj : string ref = ref "";;
let disj : string ref = ref "";;
let expoly : string ref = ref "";;
let expolyI : string ref = ref "";;
let expolyIknown : bool ref = ref false;;
let setIn : string option ref = ref None;;
let setSubeq : string option ref = ref None;;
let setPow : string option ref = ref None;;
let setimplop : string option ref = ref None;;
let setlam : string option ref = ref None;;
let replop : string option ref = ref None;;
let sepop : string option ref = ref None;;
let replsepop : string option ref = ref None;;

let setenuml : string option list ref = ref [];;
let setenumadj : string option ref = ref None;;

let seqcons : string option ref = ref None;;

let rec set_setenuml_r i h sel =
  if i > 0 then
    begin
      match sel with
      | [] -> None::set_setenuml_r (i-1) h []
      | (z::ser) -> z::set_setenuml_r (i-1) h ser
    end
  else
    begin
      match sel with
      | [] -> [Some h]
      | (_::ser) -> Some h::ser
    end

let set_setenuml_n i h =
  setenuml := set_setenuml_r i h !setenuml

let nat0 : string option ref = ref None
let natS : string option ref = ref None

(*** Hash id for polymorphic if operator: ***)
let ifop : string option ref = ref None
(*** Hash id for monomorphic if operator: ***)
let ifopset : string option ref = ref None

let megacontra : string option ref = ref None;;

let megaautosaltlimit = ref 65536;;
exception SearchLimit
exception SearchBacktrack

let mvarcnt : int ref = ref 0
let searchlim : int ref = ref 1000
let megasgdelta : (string,ptm) Hashtbl.t ref = ref (Hashtbl.create 10)

let rec local_hyp hyps g i =
  match hyps with
  | p::hypsr ->
     if conv p g !megasgdelta [] = None then
       local_hyp hypsr g (i+1)
     else
       i
  | [] -> raise Not_found

let rec tp_returns_p b a =
  if b = a then
    true
  else
    match b with
    | Ar(_,b2) -> tp_returns_p b2 a
    | _ -> false

let rec basic_megaauto lm salt cx hyps g =
  if debug then (Printf.printf "basic_megaauto %d %d %s %d %d |- %s\n" !searchlim lm (hashval_hexstring salt) (List.length cx) (List.length hyps) (tm_to_str g); flush stdout);
  if lm <= 0 || !searchlim <= 0 then raise SearchLimit;
  decr searchlim;
  try
    let i = local_hyp hyps g 0 in
    Hyp(i)
  with
  | Not_found ->
     let salt2 = hashtag salt 23l in
     let (s0,s1,s2,s3,s4,s5,s6,s7) = salt2 in
     match g with
     | Imp(p,g1) ->
        if s0 < 1610612734l then (** 7/8 of the time introduce implication **)
          let d = basic_megaauto (lm-1) salt2 cx (p::hyps) g1 in
          PLam(p,d)
        else if s1 < 1610612734l then (** 7/8 the remaining time try to apply something to prove the goal **)
          basic_megaauto_apply (lm-1) salt2 cx hyps g 0 hyps
        else (** reduce the goal to proving False **)
          let d = basic_megaauto (lm-1) salt2 cx hyps (All(Prop,DB(0))) in
          PTmAp(d,g)
     | All(Prop,DB(0)) -> (** False, just try to apply something **)
        basic_megaauto_apply (lm-1) salt2 cx hyps g 0 hyps
     | All(a,Imp(p1,g1)) when p1 = g1 -> (** covers True and refl of eq **)
        TLam(a,PLam(p1,Hyp(0)))
     | All(a,g1) ->
        if s0 < 1610612734l then (** 7/8 of the time introduce forall **)
          let d = basic_megaauto (lm-1) salt2 (a::cx) (List.map (tmshift 0 1) hyps) g1 in
          TLam(a,d)
        else if s1 < 1610612734l then (** 7/8 the remaining time try to apply something to prove the goal **)
          basic_megaauto_apply (lm-1) salt2 cx hyps g 0 hyps
        else (** reduce the goal to proving False **)
          let d = basic_megaauto (lm-1) salt2 cx hyps (All(Prop,DB(0))) in
          PTmAp(d,g)
     | _ ->
        basic_megaauto_apply (lm-1) salt2 cx hyps g 0 hyps
and basic_megaauto_apply lm salt cx hyps g i hl =
  if debug then (Printf.printf "basic_megaauto_apply %d %d %s %d %d |- %s ; %d\n" !searchlim lm (hashval_hexstring salt) (List.length cx) (List.length hyps) (tm_to_str g) i; flush stdout);
  match hl with
  | p::hr ->
     begin
       let salt2 = hashtag salt 23l in
       let (s0,s1,s2,s3,s4,s5,s6,s7) = salt2 in
       if s0 < 0l then (** 1/2 of the time skip the hyp without trying it **)
         basic_megaauto_apply lm salt2 cx hyps g (i+1) hr
       else
         try
           basic_megaauto_apply_1 lm salt2 cx hyps g (tm_to_mtm p) (fun theta -> []) (fun theta dl -> Hyp(i)) [] (fun i -> raise Not_found)
         with SearchBacktrack ->
           basic_megaauto_apply lm salt2 cx hyps g (i+1) hr
     end
  | [] ->
     raise SearchBacktrack
and basic_megaauto_apply_1 lm salt cx hyps g p subgoals d xl theta =
  if debug then (Printf.printf "basic_megaauto_apply_1 %d %d %s %d %d |- %s ; %s\n" !searchlim lm (hashval_hexstring salt) (List.length cx) (List.length hyps) (tm_to_str g) (mtm_to_str p); flush stdout);
  try
    if debug then (Printf.printf "trying to match\n"; flush stdout);
    let theta = ref (pattern_match !megasgdelta p g theta) in
    if debug then (Printf.printf "matched\n"; flush stdout);
    (** try to ground theta if it isn't **)
    List.iter
      (fun (xi,a) ->
        try
          ignore (!theta xi)
        with
          Not_found ->
          let m = basic_megaauto_tm lm salt cx a in
          let mm = tm_to_mtm m in
          let thetaold = !theta in
          theta := (fun i -> if i = xi then mm else mtm_msub (fun i -> if i = xi then mm else raise Not_found) (thetaold xi)))
      xl;
    let gl = subgoals !theta in
    let i2 = ref 0l in
    let dl = List.map (fun g -> i2 := Int32.add !i2 1l; basic_megaauto lm (hashtag salt !i2) cx hyps g) gl in
    d !theta dl
  with
  | Not_found -> raise SearchBacktrack
  | MatchFail ->
     begin
       if debug then (Printf.printf "match failed\n"; flush stdout);
       match p with
       | MAll(a,p1) ->
          let sigma = ref [] in
          for j = List.length cx - 1 downto 0 do
            sigma := MDB(j)::!sigma
          done;
          incr mvarcnt;
          let xi = !mvarcnt in
          if debug then (Printf.printf "here 0 %d\n" xi; flush stdout);
          let x = MVar(xi,!sigma) in
          if debug then (Printf.printf "here 1 %d\n" xi; flush stdout);
          let dnew = (fun theta dl -> PTmAp(d theta dl,mtm_to_tm (theta xi))) in
          if debug then (Printf.printf "here 2 p1 = %s\n" (mtm_to_str p1); flush stdout);
          basic_megaauto_apply_1 lm salt cx hyps g (mtm_ssub (x::!sigma) p1) subgoals dnew ((xi,a)::xl) theta
       | MImp(g1,p1) ->
          let subgoalsnew = (fun theta -> mtm_to_tm (mtm_msub theta g1)::subgoals theta) in
          let dnew =
            (fun theta dl ->
              match dl with
              | d1::dr -> PPfAp(d theta dr,d1)
              | _ -> raise Not_found)
          in
          basic_megaauto_apply_1 lm salt cx hyps g p1 subgoalsnew dnew xl theta
       | _ -> raise SearchBacktrack
     end
and basic_megaauto_tm lm salt cx a =
  if lm <= 0 || !searchlim <= 0 then raise SearchLimit;
  decr searchlim;
  match a with
  | Ar(a1,a2) ->
     begin
       let (s0,s1,s2,s3,s4,s5,s6,s7) = salt in
       if s7 < 0l then
         let m = basic_megaauto_tm lm (hashtag salt 23l) (a1::cx) a2 in
         Lam(a1,m)
       else
         basic_megaauto_var lm salt cx a 0 cx
     end
  | Prop ->
     let (s0,s1,s2,s3,s4,s5,s6,s7) = salt in
     if s7 < 0l then
       if s6 < 0l then
         All(Prop,DB(0))
       else
         All(Prop,Imp(DB(0),DB(0)))
     else
       basic_megaauto_apvar lm salt cx a 0 cx
  | Set -> basic_megaauto_apvar lm salt cx a 0 cx
  | _ -> raise SearchBacktrack
and basic_megaauto_apvar lm salt cx a i bl =
  match bl with
  | b::br when tp_returns_p b a ->
     let salt2 = hashtag salt 23l in
     let (s0,s1,s2,s3,s4,s5,s6,s7) = salt2 in
     if s7 < 0l then
       begin
         try
           basic_megaauto_spine (lm-1) salt2 cx a b (DB(i))
         with SearchBacktrack ->
           basic_megaauto_apvar lm salt cx a (i+1) br
       end
     else
       basic_megaauto_apvar lm salt cx a (i+1) br
  | _::br -> basic_megaauto_apvar lm salt cx a (i+1) br
  | [] -> raise SearchBacktrack
and basic_megaauto_var lm salt cx a i bl =
  match bl with
  | b::br when a = b ->
     let salt2 = hashtag salt 23l in
     let (s0,s1,s2,s3,s4,s5,s6,s7) = salt2 in
     if s7 < 0l then
       DB(i)
     else
       basic_megaauto_var lm salt cx a (i+1) br
  | _::br -> basic_megaauto_var lm salt cx a (i+1) br
  | [] -> raise SearchBacktrack
and basic_megaauto_spine lm salt cx a b m =
  if lm <= 0 || !searchlim <= 0 then raise SearchLimit;
  decr searchlim;
  if a = b then
    m
  else
    match b with
    | Ar(b1,b2) ->
       let m1 = basic_megaauto_tm (lm-1) salt cx b1 in
       basic_megaauto_spine lm salt cx a b2 (Ap(m,m1))
    | _ -> raise SearchBacktrack

let rec basic_megaauto_search masl salt cx hyps g =
  if masl <= 0 then (Printf.printf "Line %d Char %d: Failed to find salt\n" !lineno !charno; flush stdout; raise SearchLimit);
  try
    mvarcnt := 0; searchlim := 1000;
    if debug then Printf.printf "Trying salt %s\n" (md256_hexstring salt); flush stdout;
    let d = basic_megaauto 50 salt cx hyps g in
    let h = md256_hexstring salt in
    let n = ref 63 in
    while (!n >= 0 && h.[!n] = '0') do
      decr n
    done;
    Printf.printf "Line %d Char %d: Use ?%s?\n" !lineno !charno (String.sub h 0 (!n+1)); flush stdout;
    d
  with
  | SearchLimit -> basic_megaauto_search_nxt (masl-1) salt cx hyps g
  | SearchBacktrack -> basic_megaauto_search_nxt (masl-1) salt cx hyps g
and basic_megaauto_search_nxt masl salt cx hyps g =
  let (s0,s1,s2,s3,s4,s5,s6,s7) = salt in
  if (s1,s2,s3,s4,s5,s6,s7) = (0l,0l,0l,0l,0l,0l,0l) then
    let s0r = Hashaux.int32_rev s0 in
    if s0r = -1l then
      basic_megaauto_search masl (0l,1l,0l,0l,0l,0l,0l,0l) cx hyps g
    else
      basic_megaauto_search masl (Hashaux.int32_rev (Int32.add s0r 1l),0l,0l,0l,0l,0l,0l,0l) cx hyps g
  else
    if s0 = -1l then
      if s1 = -1l then
        if s2 = -1l then
          basic_megaauto_search masl (s0,s1,s2,Int32.add s3 1l,s4,s5,s6,s7) cx hyps g
        else
          basic_megaauto_search masl (s0,s1,Int32.add s2 1l,s3,s4,s5,s6,s7) cx hyps g
      else
        basic_megaauto_search masl (s0,Int32.add s1 1l,s2,s3,s4,s5,s6,s7) cx hyps g
    else
      basic_megaauto_search masl (Int32.add s0 1l,s1,s2,s3,s4,s5,s6,s7) cx hyps g

let rec basic_megaauto_small_search i j k l cx hyps g =
  let trysalt s1 =
    try
      mvarcnt := 0; searchlim := 1000;
      let d = basic_megaauto 50 (s1,0l,0l,0l,0l,0l,0l,0l) cx hyps g in
      let h = md256_hexstring (s1,0l,0l,0l,0l,0l,0l,0l) in
      let n = ref 63 in
      while (!n >= 0 && h.[!n] = '0') do
        decr n
      done;
      Printf.printf "Line %d Char %d: Use ?%s?\n" !lineno !charno (String.sub h 0 (!n+1)); flush stdout;
      d
    with SearchBacktrack -> raise SearchLimit
  in
  try
    trysalt (Int32.logor (Int32.shift_left i 28)
               (Int32.logor (Int32.shift_left j 24)
                  (Int32.logor (Int32.shift_left k 20)
                     (Int32.shift_left l 16))))
  with SearchLimit ->
        if i < 15l then
          basic_megaauto_small_search (Int32.add i 1l) j k l cx hyps g
        else if j < 15l then
          basic_megaauto_small_search 0l (Int32.add j 1l) k l cx hyps g
        else if k < 15l then
          basic_megaauto_small_search 0l 0l (Int32.add k 1l) l cx hyps g
        else if l < 15l then
          basic_megaauto_small_search 0l 0l 0l (Int32.add l 1l) cx hyps g
        else
          raise SearchLimit

let megaauto pre ops cx hyps g =
  if pre = "" then
    if not (ops = "") && ops.[0] = '*' then (** search for an appropriate salt **)
      let n = String.length ops - 1 in
      let h =
        if n = 64 then
          String.sub ops 1 64
        else if n > 64 then
          raise (Failure (Printf.sprintf "salt for megaauto too long \"%s\"" ops))
        else
          let sbuf = Buffer.create 64 in
          for i = 1 to n do Buffer.add_char sbuf (ops.[i]) done;
          for i = n to 63 do Buffer.add_char sbuf '0' done;
          Buffer.contents sbuf
      in
      let salt = hexstring_md256 h in
      if debug then (Printf.printf "Line %d Char %d: Trying Basic Megaauto Search (will likely hang until successful or killed)\n" !lineno !charno; flush stdout);
      basic_megaauto_search !megaautosaltlimit salt cx hyps g
    else if ops = "+" then (** search for an appropriate small salt **)
      begin
        try
          basic_megaauto_small_search 0l 0l 0l 0l cx hyps g
        with SearchLimit ->
          Printf.printf "Line %d Char %d: Failed to find small salt\n" !lineno !charno; flush stdout;
          raise SearchLimit
      end
    else
      let n = String.length ops in
      let h =
        if n = 64 then
          ops
        else if n > 64 then
          raise (Failure (Printf.sprintf "salt for megaauto too long \"%s\"" ops))
        else
          let sbuf = Buffer.create 64 in
          Buffer.add_string sbuf ops;
          for i = n to 63 do Buffer.add_char sbuf '0' done;
          Buffer.contents sbuf
      in
      let salt = hexstring_md256 h in
      mvarcnt := 0; searchlim := 1000;
      basic_megaauto 50 salt cx hyps g
  else
    raise (Failure (Printf.sprintf "unknown prefix \"%s\" for megaauto" pre))
